﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace EPHConsole
{
    class Program
    {
        static void Main(string[] args)
        {
            if (args.Length != 1)
            {
                Console.WriteLine("Usage: EPHConsole <ballot file>");
                Console.WriteLine("Ballot file should be plain text, one ballot per line.");
                Console.WriteLine("Ballots should consist of integers (representing works) seperated by spaces:");
                Console.WriteLine("1 2 3 4 5");
                Console.WriteLine("Press enter to exit.");
                Console.ReadLine();
                return;
            }

            System.IO.StreamReader file;
            try
            {
                file = new System.IO.StreamReader(args[0]);
            }
            catch(Exception e) {
                Console.WriteLine("Couldn't open ballot file.");
                Console.WriteLine("Press enter to exit.");
                Console.ReadLine();
                return;
            }
            String line;
            String[] linesplit;
            List<Ballot> AllBallots = new List<Ballot>();
            while ((line = file.ReadLine()) != null)
            {
                linesplit = line.Split(' ');
                try
                {
                    AllBallots.Add(new Ballot(linesplit));
                }
                catch(Exception e)
                {
                    Console.WriteLine("Bad data in ballot file.");
                    Console.WriteLine("Press enter to exit.");
                    Console.ReadLine();
                    return;
                }
                
            }

            Dictionary<int, SFResult> results = new Dictionary<int, SFResult>();

            //Test to display the contents of all the ballots.

            foreach (Ballot b in AllBallots)
            {
                Console.WriteLine(b.display());
            }

            Console.WriteLine();
            Console.ReadLine();

            bool complete = false;
            int round = 0;
            float least;
            List<KeyValuePair<int, SFResult>> roster;
            List<KeyValuePair<int, SFResult>> selectedWorks = new List<KeyValuePair<int, SFResult>>();
            while (!complete) //loop of selection and elimination until we're done
            {
                round++;
                Console.WriteLine();
                Console.WriteLine("Round " + round);

                CountBallots(AllBallots, results); //generate correct point and vote totals
                roster = results.ToList();
                roster.Sort(delegate(KeyValuePair<int, SFResult> x, KeyValuePair<int, SFResult> y)
                {
                    return (x.Value.points.CompareTo(y.Value.points));
                }); //Roster is now the set of works, sorted by points

                //Display sorted by points
                foreach (KeyValuePair<int, SFResult> k in roster)
                {
                    Console.WriteLine(k.Key + " " + k.Value.points + "/" + k.Value.votes);
                }

                least = roster[0].Value.points; //the lowest point score
                if (roster[1].Value.points > least) //if this is true, only one work with least value
                {
                    least = roster[1].Value.points; //threshold is now second-least
                }

                Console.WriteLine("Threshhold: " + least);


                selectedWorks.Clear();
                foreach (KeyValuePair<int, SFResult> k in roster) //add all works at threshold or below to selection
                {
                    if (k.Value.points <= least) selectedWorks.Add(k); //if a work was selected, add it along with it's vote count
                }

                selectedWorks.Sort(delegate(KeyValuePair<int, SFResult> x, KeyValuePair<int, SFResult> y)
                {
                    return (x.Value.votes.CompareTo(y.Value.votes));
                }); //selectedWorks is now sorted by voteCount


                int lowvote = selectedWorks[0].Value.votes; //the lowest vote total
                int counter = selectedWorks.Count - 1; //start from the top
                while (counter > 0) //no need to check the bottom one
                {
                    if (selectedWorks[counter].Value.votes > lowvote) selectedWorks.RemoveAt(counter); //throw away all that are not tied for the lowest vote total
                    counter--;
                }

                //Now everything in selectedWorks is tied on votecount
                //By definition there are only two point totals at most -- either the lowest or the second-lowest
                //And if there are two, the the lowest is only a single work
                bool eliminateAll = true; //true if we should remove everything from selectedWorks, false if only one work
                if (selectedWorks.Count > 1) //there's at least two things left
                {
                    if (selectedWorks[0].Value.points < selectedWorks[1].Value.points) eliminateAll = false; //lowest is single work
                }

                int removeCount = 1;
                if (eliminateAll) removeCount = selectedWorks.Count; //the number of works we're removing, to check if we're done

                int remainingWorks = roster.Count - removeCount;
                if (remainingWorks >= 5) //we can eliminate all the tied works.
                {
                    if (remainingWorks == 5) complete = true; //elimination brings us to exactly 5, and we're done.

                    if (eliminateAll) //eliminate everything in SelectedWorks
                    {
                        counter = 0;
                        while (counter < selectedWorks.Count)
                        {
                            RemoveWork(AllBallots, selectedWorks[counter].Key);
                            Console.WriteLine("Eliminating " + selectedWorks[counter].Key + ".");
                            counter++;
                        }

                    }
                    else //eliminate only 1
                    {
                        RemoveWork(AllBallots, selectedWorks[0].Key);
                        Console.WriteLine("Eliminating " + selectedWorks[0].Key + ".");
                    }

                }
                else //eliminating all tied works would bring total below 5, so keep them and finish
                {
                    complete = true;
                }

            }

            //Final Display
            Console.WriteLine();
            Console.WriteLine("Final Results:");
            CountBallots(AllBallots, results); //generate correct point and vote totals
            roster = results.ToList();
            roster.Sort(delegate(KeyValuePair<int, SFResult> x, KeyValuePair<int, SFResult> y)
            {
                return (x.Value.points.CompareTo(y.Value.points));
            }); //Roster is now the set of works, sorted by points
            //Display sorted by points
            foreach (KeyValuePair<int, SFResult> k in roster)
            {
                Console.WriteLine(k.Key + " " + k.Value.points + "/" + k.Value.votes);
            }

            Console.ReadLine();
        }

        public static void CountBallots(List<Ballot> theBallots, Dictionary<int, SFResult> results)
        {
            float pointValue;
            results.Clear();

            foreach (Ballot b in theBallots)
            {
                if (b.works.Count() > 0) //ignore empty ballots
                {
                    pointValue = 1.0f / b.works.Count();
                    foreach (int w in b.works)
                    {
                        if (results.ContainsKey(w))
                        {
                            results[w].points += pointValue;
                            results[w].votes += 1;
                        }
                        else
                        {
                            results.Add(w, new SFResult(pointValue, 1));
                        }
                    }
                }
            }
        }

        public static void RemoveWork(List<Ballot> theBallots, int theWork)
        {
            foreach (Ballot b in theBallots)
            {
                b.works.Remove(theWork);
            }
        }
    }

         

    public class SFResult
    {
        public float points;
        public int votes;

        public SFResult()
        {
            points = 0.0f;
            votes = 0;
        }

        public SFResult(float p, int v)
        {
            points = p;
            votes = v;
        }
    }

    public class Ballot
    {

        public String display()
        {
            String res = "";
            foreach (int x in works)
            {
                res += x + "\t";
            }
            return res;
        }

        public Ballot()
        {
            works = new List<int>();
        }

        public Ballot(string[] incWork)
        {
            works = new List<int>();

            foreach (string s in incWork)
            {
                if (s != "") works.Add(Int32.Parse(s));
            }
        }

        public Ballot(int a)
        {
            works = new List<int>();

            works.Add(a);
        }

        public Ballot(int a, int b)
        {
            works = new List<int>();

            works.Add(a);
            works.Add(b);
        }

        public Ballot(int a, int b, int c)
        {
            works = new List<int>();

            works.Add(a);
            works.Add(b);
            works.Add(c);
        }

        public Ballot(int a, int b, int c, int d)
        {
            works = new List<int>();

            works.Add(a);
            works.Add(b);
            works.Add(c);
            works.Add(d);
        }

        public Ballot(int a, int b, int c, int d, int e)
        {
            works = new List<int>();

            works.Add(a);
            works.Add(b);
            works.Add(c);
            works.Add(d);
            works.Add(e);
        }
        public List<int> works;
    }

}
